import pyvisa
import numpy as np
from time import sleep
# import pkg_resources
import os

"""
Utility
"""

# scripts = {
#     "buffer_reset": pkg_resources.resource_filename("cpdctrl", "keithley_scripts/buffer_reset.lua"),
#     "instrument_reset":   pkg_resources.resource_filename("cpdctrl", "keithley_scripts/smua_reset.lua"),
# }
scripts = {

    "instrument_reset": "~/cpd-dev/cpdctrl/cpdctrl/keithley_scripts/reset.scpi",
}



class Keithley2700:
    """
    Wrapper class for the Keithley2700 SMU controlled via pyvisa
    """
    
    def __init__(self, visa_backend="", check_front_switch=True):
        rm = pyvisa.ResourceManager(visa_backend)
        resources = rm.list_resources()
        if len(resources) < 1:
            raise Exception("No resources found.")
        elif len(resources) == 1:
            print(f"Opening the only resource found: {resources[0]}")
            self.instr = rm.open_resource(resources[0])
        elif len(resources) > 1:
            print(f"Resources:")
            for i, r in enumerate(resources):
                print(f"{i:02}: {r}")
            instr = int(input("Select an instrument: "))
            self.instr = rm.open_resource(resources[instr])
        if check_front_switch:
            self._check_front_input_selected()
                
    def __del__(self):
        """Properly close the instrument connection"""
        self.instr.close()
        
    def _check_front_input_selected(self):
        """
        Make sure the front switch selecting the inputs selects the FRONT inputs
        Raises
        ------
        Exception
            If front input state != 1

        Returns
        -------
        switch : str
            Front input state
        """
        switch = self.instr.query("SYSTem:FRSwitch?").strip("\n")
        if switch != "1":
            raise Exception("The Keithley's INPUT switch must select the [F]ront inputs")
        return switch
        
    def run(self, code, verbose=False):
        """
        Run SCPI code on the device by writing it.
        Empty lines, leading whitespaces and lines starting with ' or # are ignored.
        Parameters
        ----------
        code : str
            SCPI commands
        """
        script = '\n'.join([l.strip(" ") for l in code.strip(" ").strip("\n").split("\n") if len(l) > 0 and l[0] not in "#'"])
        if verbose:
            print(f"Running code:\n{script}")
        self.instr.write(script)
        
    def run_script(self, script_path, verbose=False):       
        """
        Run a script from the host on the instrument
        @param script_path : full path to the script
        
        Parameters
        ----------
        script_path : str
            full path to the script
        verbose : bool, optional
            If true, print a message when the script is run. The default is False.

        Returns
        -------
        None.
        """
        with open(os.path.expanduser(script_path), "r") as file:
            script = file.read()
        
        if verbose: print(f"Running script: {script_path}")
        self.run(script)
    
    def beep(self, length=0.5, pitch=1000):
        """The Keithley2700 cant beep on command :("""
        pass


    
    def reset(self, verbose=False):
        """
        Reset smua and its buffers
        @param instr : pyvisa instrument
        """
        self.run_script(scripts["instrument_reset"], verbose=verbose)
        self.buffer_reset()
    
    def buffer_reset(self):
        buffer_reset = """
        TRACe:CLEar
        TRACe:CLEar:AUTO ON
        SYSTem:TSTamp:TYPE RELative
        """
        self.run(buffer_reset)
        
    def buffer_get_size(self, buffer_nr=1):
        n = self.instr.query(f"TRACe:POINts?").strip("\n")
        return int(float(n))
    
    def buffer_set_size(s):
        if not type(s) == int or s < 2 or s > 55000:
            raise ValueError(f"Invalid buffer size: {s}. Must be int and between 2 and 55000")
        self.instr.write(f"TRACe:POINts {s}")
        
    def buffer_collect(self, verbose=False):
        """
        Get the buffer as 2D - np.array
        @param buffer_nr : 1 or 2, for smua.nvbuffer1 or 2
        @returns 2D numpy array:
            i - ith reading:
                0: timestamps
                1: readings
        """
        readings = self.instr.query("TRACe:DATA?")
        tdata = []
        vdata = []
        for reading in readings.split("\n"):
            t, v = self.process_reading(reading)
            tdata.append(t)
            vdata.append(v)
            
        if verbose:
            print(f"readings from buffer:\n{vdata}\ntimestamps:\n{tdata}")
        buffer = np.vstack((tdata, vdata)).T
        return buffer
    
    
    def collect_buffer_range(self, range_=(1, -1), buffer_nr=1, verbose=False):
        """
        Get the buffer as 2D - np.array
        @param instr : pyvisa instrument
        @param buffer_nr : 1 or 2, for smua.nvbuffer1 or 2
        @returns 2D numpy array:
            i - ith reading:
                0: timestamps
                1: readings
        """
        buffername = get_buffer_name(buffer_nr)
        # instr.write("format.data = format.DREAL\nformat.byteorder = format.LITTLEENDIAN")
        # buffer = instr.query_binary_values(f"printbuffer(1, {buffername}.n, {buffername})", datatype='d', container=np.array)
        if range_[1] == -1:
            range_ = (range_[0], f"{buffername}.n")
        instr.write("format.data = format.ASCII\nformat.asciiprecision = 7")
        timestamps = instr.query_ascii_values(f"printbuffer({range_[0]}, {range_[1]}, {buffername}.timestamps)", container=np.array)
        readings = instr.query_ascii_values(f"printbuffer({range_[0]}, {range_[1]}, {buffername}.readings)", container=np.array)
        if verbose:
            print(f"readings from {buffername}: {readings}, \ntimestamps: {timestamps}")
        buffer = np.vstack((timestamps, readings)).T
        return buffer
    
    def process_reading(self, reading: str):
        """
        process a reading. Only works with VDC and relative time stamps right now!
        '-1.19655066E+01VDC,+9627.275SECS,+64993RDNG#\n'
        """
        parts = reading.split(",")
        if len(parts) != 3:
            raise ValueError(f"Invalid reading: '{reading}'")
        vdc = float(parts[0][:-3])
        timestamp = float(parts[1][:-4])
        # TODO last value
        return timestamp, vdc        
        
    def read_value(self):
        reading = self.instr.query("READ?")
        return self.process_reading(reading)
    
    def measure(self, interval, update_func=None, max_measurements=None):
        """
        @details:
            - Resets the buffers
            - Until KeyboardInterrupt:
                - Take measurement
                - Call update_func
                - Wait interval
            Uses python's time.sleep() for waiting the interval, which is not very precise. Use measure_count for better precision
            You can take the data from the buffer afterwards, using save_csv
        @param instr: pyvisa instrument
        @param update_func: Callable that processes the measurements: (index, ival, vval) -> None
        @param max_measurements : maximum number of measurements. None means infinite
        """
        self.reset(verbose=True)
        try:
            i = 0
            while max_measurements is None or i < max_measurements:
                tval, vval = self.read_value()
                if update_func:
                    update_func(i, tval, vval)
                sleep(interval)
                i += 1
        except KeyboardInterrupt:
            pass
        # instr.write("smua.source.output = smua.OUTPUT_OFF")
        print("Measurement stopped" + " "*50)

def init():
    return Keithley2700()
